(declare (export polys-string lex-rational lex-rational-xyz buchberger g-reduce
  gbasis grobner? poly-string lex rlex grevlex rational-field m-posdiv?))
;;
;; Groebner Basis
;; By Naruto Canada
;; 
;; Read more about Groebner Basis here:
;; http://www.geocities.com/famancin/
;;
;; With code snippets from:
;; comp.lang.scheme
;; http://mitpress.mit.edu/sicp/full-text/book/book.html
;; http://www.ccs.neu.edu/home/dorai/t-y-scheme/t-y-scheme.html
;; http://www-swiss.ai.mit.edu/~jaffer/SCM.html
;; bit-scheme
;;

(use numbers)

(define call/ec call/cc)
(define (print x) (display x) (newline))
(define (identity x) x)
(define (map-print m) (map print m))
(define (negate x) (* -1 x))


;; flatten one level 
(define (flatten-1 lst)
 (define ret '())
 (for-each (lambda (a) (if (list? a) (set! ret (append ret a)) 
                                     (set! ret (append ret (list a))) 
                       )
 ) lst)
 ret
)


;; flatten a list completely
(define (flatten z)
  (define y '())
  (define (traverse x)
     (cond ((null? x))
        ((not (pair? x)) (set! y (append y (list x))))
        (else (begin (traverse (car x))
                     (traverse (cdr x))))))
  (traverse z) 
  y
)


(define (assign x y) (set-car! x y) y)
(define (xor x y) (or (and x (not y)) (and (not x) y)))


(define (first x) (car x))
(define (second x) (cadr x))
(define (third x) (caddr x))
(define (fourth x) (cadddr x))
(define (fifth x) (car (cddddr x)))
(define (last lst) 
  (if (null? (cdr lst)) (car lst)
      (last (cdr lst))
  )
)


(define (for-loop x x-max do-stuff) 
  (if (<= x x-max) 
      (begin (do-stuff x) (for-loop (+ x 1) x-max do-stuff))
  )
)


(define (foldr1 binop l)
  (if (null? (cdr l)) (car l)
      (binop (car l) (foldr1 binop (cdr l)))
  )
)
(define (foldl1 binop l)
    (if (null? (cdr l)) (car l)
        (foldl1 binop (cons (binop (car l) (cadr l)) (cddr l)))
    )
)


(define (unfoldr1 binop lst)
  (if (null? lst) lst
      (if (null? (cdr lst)) (car lst) 
          (cons binop (list (car lst) (unfoldr1 binop (cdr lst))))
      )
  )
)


(define (unfoldl1 binop lst)
  (unfoldr1 binop (reverse lst))
)


(define (square x) (* x x))
(define (square? n) (exact? (sqrt n)))


(define (filter predicate sequence)
  (if (null? sequence) sequence
      (if (predicate (car sequence))
          (cons (car sequence) (filter predicate (cdr sequence)))
          (filter predicate (cdr sequence))
      )
  )
)


(define (interval a b) (if (<= a b) (cons a (interval (+ a 1) b)) '()) )


(define (make-list n x) 
  (define ret '())
  (for-loop 1 n (lambda (i) 
    (set! ret (cons x ret))
  ))
  ret
)


(define (mod a n) 
  (if (list? a) (map (lambda (x) (mod x n)) a)
    (if (>= a 0) (remainder a n) 
      (remainder (+ (remainder a n) n) n)
    )
  )
)


(define (prime? x)
  (if (< x 2) #f
    (call/ec (lambda (return)
      (for-loop 2 (- x 1) (lambda (i)
         (if (= (remainder x i) 0) (return #f))
      ))
    #t))
  )
)


;; return 1 if x^2==a mod b exists 
;; else return -1
(define (quadres a b)
  (call/ec (lambda (return)
    (let ((aa (mod a b)))
         (for-each (lambda (x) (if (= aa (mod (square x) b)) (return 1))) (interval 1 b))
          -1
    )
  ))
)


(define-macro (while test body1 . bodyrest)
 `(
   (lambda (test-thunk body-thunk)
    (letrec
      ((loop (lambda (prev-result)
               (if (test-thunk) (loop (body-thunk)) prev-result))))
      (loop #f)))
  (lambda () ,test)             ; test-thunk
  (lambda () ,body1 ,@bodyrest) ; body-thunk
)) 


;; 
;; http://www.ccs.neu.edu/home/dorai/t-y-scheme/t-y-scheme-Z-H-1.html
;;
(define amb-fail '*)
(define initialize-amb-fail
  (lambda ()
    (set! amb-fail
      (lambda ()
        (error "amb tree exhausted")))))
(define-macro amb
  (lambda alts...
    `(let ((+prev-amb-fail amb-fail))
       (call/ec
        (lambda (+sk)
          ,@(map (lambda (alt)
                   `(call/cc
                     (lambda (+fk)
                       (set! amb-fail
                         (lambda ()
                           (set! amb-fail +prev-amb-fail)
                           (+fk 'fail)))
                       (+sk ,alt))))
                 alts...)
          (+prev-amb-fail))))))
(define number-between
  (lambda (lo hi)
    (let loop ((i lo))
      (if (> i hi) (amb)
          (amb i (loop (+ i 1)))))))
(define assert
  (lambda (pred)
    (if (not pred) (amb))))
(define-macro bag-of
  (lambda (e)
    `(let ((+prev-amb-fail amb-fail)
           (+results '()))
       (if (call/ec
            (lambda (+k)
              (set! amb-fail (lambda () (+k #f)))
              (let ((+v ,e))
                (set! +results (cons +v +results))
                (+k #t))))
           (amb-fail))
       (set! amb-fail +prev-amb-fail)
       (reverse +results))))
(define-macro amb-run
  (lambda (e)
    `(let ((+prev-amb-fail amb-fail)
           (+results 0))
       (if (call/ec
            (lambda (+k)
              (set! amb-fail (lambda () (+k #f)))
              (let ((+v ,e))
                (set! +results (+ 1 +results))
                (+k #t))))
           (amb-fail))
       (set! amb-fail +prev-amb-fail)
       (print (string-append "# amb-run: " (number->string +results)))  
     )))

(define gen-prime
  (lambda (n)
    (let ((i (number-between 2 n)))
      (assert (prime? i))
      i)))

(define amb-list
  (lambda (lst)
    (let loop ((x lst))
      (if (null? x) (amb)
          (amb (car x) (loop (cdr x)))
      ))))



(define (AND x y) (if (and x y) #t #f))
(define (OR x y) (if (or x y) #t #f))

;; takes 0 1 -> #f #t
(define (list->boolean lst)
  (map (lambda (n)
    (if (= n 0) #f 
      (if (= n 1) #t (error "only takes 0 or 1"))
    )
  ) lst)
)


;; http://mitpress.mit.edu/sicp/full-text/book/book.html
(define (memoize f)
  (define (lookup key table)
    (let ((record (assoc key (cdr table))))
      (if record (cdr record) #f)))
  (define (insert! key value table)
    (let ((record (assoc key (cdr table))))
      (if record
          (set-cdr! record value)
          (set-cdr! table
                    (cons (cons key value) (cdr table)))))
    'ok)
  (define (make-table) (list '*table*))
  (let ((table (make-table))) 
    (lambda (x)
      (let ((previously-computed-result (lookup x table)))
        (or previously-computed-result
            (let ((result (f x)))
              (insert! x result table)
              result))))))
(define memo-fib
  (memoize (lambda (n) 
             (cond ((= n 0) 0)
                   ((= n 1) 1)
                   (else (+ (memo-fib (- n 1))
                            (memo-fib (- n 2))))))))


;; http://www.cs.bgu.ac.il/~elhadad/scheme/merge.html
;; http://www.cs.bgu.ac.il/~elhadad/scheme/insertion.html
(define (insertion-sort < ls)
  (define (reverse-it ls acc) 
    (if (null? ls) acc (reverse-it (cdr ls) (cons (car ls) acc)))
  )
  (define (insert new sorted)
    (let loop ((rest sorted) (passed '()))
      (cond ((null? rest)       (reverse-it passed (list new)))
            ((< new (car rest)) (reverse-it passed (cons new rest)))
            (else (loop (cdr rest) (cons (car rest) passed)))
      )
    )
  )
  (let outer-loop ((remaining ls) (done '()))
    (if (null? remaining) done
        (outer-loop (cdr remaining) (insert (car remaining) done))
    )
  )
)


(define (find-inverse-mod-p a p)
  (call/ec (lambda (return)
    (for-loop 1 (- p 1) (lambda (i)
       (if (= (mod (* a i) p) 1) (return i))
    )) 
  (error "no inverse mod p")))  
)


(define (list-copy ls) 
  (if (pair? ls)
      (cons (list-copy (car ls)) (list-copy (cdr ls)))
      ls
  )
)


;; lst must have at least one element already
(define (cons! x lst)
  (define ret (list-copy lst))
  (set-car! lst x)
  (set-cdr! lst ret)
  lst
)


;; lst must have at least one element already
(define (append! lst1 lst2)
  (if (pair? (cdr lst1)) (append! (cdr lst1) lst2) (begin
      (set-cdr! lst1 lst2) 
  ))
  lst1
)


(define (nth-cdr n lst)
  (if (> n 0) (nth-cdr (- n 1) (cdr lst)) lst)
)


(define (take n lst)
   (if (<= n 0) '()
       (if (< n (length lst))
           (reverse (nth-cdr (- (length lst) n) (reverse lst)))
           lst
       ) 
   )
)


(define (iota count start inc)
  (if (<= count 0) '()
      (cons start (iota (- count 1) (+ start inc) inc))
  )
)


(define (take-right n lst) (nth-cdr (- (length lst) n) lst))


(define (delete = x ls)
  (define ret '())
  (for-each (lambda (y) 
    (if (not (= x y)) (set! ret (append ret (list y))) )
  ) ls)
  ret  
)


(define (contains? = element lst)
  (if (null? lst) #f
      (if (= element (car lst)) #t
          (contains? = element (cdr lst))
      )
  )
) 


(define (search-list i = element lst)
  (if (null? lst) -1
      (if (= element (car lst)) i
          (search-list (+ i 1) = element (cdr lst))
      )
  )
)


(define (list->set = lst) 
  (if (null? lst) lst
      (cons (car lst) (list->set = (delete = (car lst) (cdr lst))))
  )
)


;; delete elements from set1 using elements from set2
(define (set-minus equal? set1 set2)
  (define result (list->set equal? set1))
  (map (lambda (x) (set! result (delete equal? x result))) set2)
  result
)


;; equal? is for set element comparison
(define (set-cmp? equal? s1 s2) 
  (and (null? (set-minus equal? s1 s2))
       (null? (set-minus equal? s2 s1))
  )
  ;(null? (set-minus equal? s1 s2))
)


(define (convert-to-base n b)
  (if (= 0 n) '()
    (cons (remainder n b) (convert-to-base (quotient n b) b))
  )
)


(define (convert-from-base n b)
 (define total 0)
 (define d 0)
 (let iter ((s n))
      (if (not (null? s)) 
          (begin 
                (if (not (zero? (car s))) (set! total (+ total (* (car s) (expt b d)))))
                (set! d (+ d 1))
                (iter (cdr s))
          )
      )     
 ) 
 total
)


;; append zeroes to a list making it length n 
(define (list->row n lst) (append lst (iota (- n (length lst)) 0 0)))


(define (idx-interval lst) (interval 0 (- (length lst) 1)))


(define (swap i j lst)
  (define x (list-ref lst i))
  (define y (list-ref lst j))
  (map (lambda (z k)
    (if (= k i) y (if (= k j) x z))
  ) lst (idx-interval lst) )
)


(define (swap! i j lst)
  (define x (list-ref lst i))
  (define y (list-ref lst j))
  ;(map (lambda (z k)
  ;  (if (= k i) (set! z y) (if (= k j) (set! z x) z))
  ;) lst (idx-interval lst) )
  (set-car! (nth-cdr j lst) x) 
  (set-car! (nth-cdr i lst) y) 
  lst
)


(define (list-set! lst i x)
  (set-car! (nth-cdr i lst) x) 
  lst
)


;; map a list to a function then sort
;; return list of pairs with original location (element . location)
(define (map-sort func < lst) 
  (insertion-sort (lambda (x y) (< (car x) (car y)))
    (map (lambda (x i) (cons (func x) i))  lst (idx-interval lst) )
  )
)


(define (nth-apply n func x)
  (define ret x)
  (for-loop 1 n (lambda (i)
    (set! ret (func ret))
  ))
  ret
)


(define (rotate-right lst)
  (if (null? lst) lst
    (cons (last lst) (take (- (length lst) 1) lst))
  )
)


(define (rotate-left ls)  (append (cdr ls) (list (car ls))))


(define (rotate-left-n n lst) (nth-apply n rotate-left lst))
(define (rotate-right-n n lst) (nth-apply n rotate-right lst))


;; look no need for call/ec
(define (all-zero? zero? lst)
  (if (null? lst) #t 
      (if (not (zero? (car lst))) #f 
          (all-zero? zero? (cdr lst))
      )
  )
)


(define (all-zero-matrix? zero? lst)
  (if (null? lst) #t 
      (if (not (all-zero? zero? (car lst))) #f 
          (all-zero-matrix? zero? (cdr lst))
      )
  ) 
)


(define (identity-matrix n)
  (define (ith-one i n) (map (lambda (x) (if (= x i) 1 0)) (interval 1 n)))
  (map (lambda (i) (ith-one i n)) (interval 1 n))
)


(define (rational-field op)
  (cond ((eq? op 'zero?)                       zero? )
        ((eq? op 'one?)        (lambda (x) (= 1 x))  )
        ((eq? op 'add)                             + )
        ((eq? op 'mul)                             * )
        ((eq? op 'mul-inv)     (lambda (x) (/  1 x)) )
        ((eq? op 'negate)      (lambda (x) (* -1 x)) )
        ((eq? op 'zero)                            0 )
        ((eq? op 'one)                             1 )
        ((eq? op 'greater-than)                    > )
        ((eq? op 'abs)                           abs )
        ((eq? op 'str)                number->string )
        (else                   (error "no such op") )
  )
)

(define (Zp-field p)
  (lambda (op)
    (cond ((eq? op 'zero?)        (lambda (x)   (= 0 (remainder x p)))    )
          ((eq? op 'one?)         (lambda (x)   (= 1 (mod x p)))          )
          ((eq? op 'add)          (lambda (x y) (mod (+ x y) p))          )
          ((eq? op 'mul)          (lambda (x y) (mod (* x y) p))          )
          ((eq? op 'mul-inv)      (lambda (x)   (find-inverse-mod-p x p)) )
          ((eq? op 'negate)       (lambda (x)   (mod (* -1 x) p))         )
          ((eq? op 'zero)                                               0 )
          ((eq? op 'one)                                                1 )
          ((eq? op 'greater-than) (lambda (x y) (> (mod x p) (mod y p)))  )
          ((eq? op 'abs)          (lambda (x) (mod x p))                  )
          ((eq? op 'str)          (lambda (x) (number->string (mod x p))) )
          (else                                      (error "no such op") )
    )
  )
)


(define (transpose matrix)
  (map (lambda (j) 
    (map (lambda (row) (list-ref row j)) matrix)
  ) (interval 0 (- (length (car matrix)) 1)) )
) 


(define factorial (memoize (lambda (n) 
  (if (<= n 1) 1 (* n (factorial (- n 1))))
)))


(define (cross-product lst)
  (define (append-map proc first . rest) (apply append (apply map proc first rest)))
  (if (null? lst) (list '())
      (append-map 
        (lambda (x)
          (map (lambda (prev-result) (cons x prev-result))
                 (cross-product (cdr lst))))
        (car lst))
  )
) 



(define (matrix-ref matrix row col)
  (list-ref (list-ref matrix row) col)
)


(define (matrix-set! matrix row col value)
  (set-car! (nth-cdr col (list-ref matrix row)) value) 
  matrix
) 


(define (matrix-swap-columns! matrix i j)
  (define m (transpose (swap! i j (transpose matrix))) )
  (for-each (lambda (i)
    (list-set! matrix i (list-ref m i))
  ) (idx-interval matrix) )
  matrix
)


;; only works for square matrix
(define (transpose! matrix)
  (define m (transpose matrix))
  (for-each (lambda (i)
    (list-set! matrix i (list-ref m i))
  ) (idx-interval matrix) )
  matrix
)


(define (matrix-row-op! matrix i func)
  (list-set! matrix i (map func (list-ref matrix i)))
  matrix
)


(define (matrix-column-op! matrix j func)
  (for-each (lambda (row)
    (list-set! row j (func (list-ref row j)))
  ) matrix)
  matrix 
)


(define (matrix-get-column matrix j)
  (map (lambda (row) (list-ref row j)) matrix)
)


(define (matrix-set-column! matrix j lst)
  (for-each (lambda (row x)
    (list-set! row j x)
  ) matrix lst)
  matrix
)


(define (matrix-add field m1 m2)
  (map (lambda (row1 row2)
    (map (lambda (x y)
      ((field 'add) x y)
    ) row1 row2)
  ) m1 m2)
)


(define (matrix-sub field m1 m2)
  (map (lambda (row1 row2)
    (map (lambda (x y)
      ((field 'add) x ((field 'negate) y))
    ) row1 row2)
  ) m1 m2)
)


(define (inner-product field lst1 lst2)
  (foldr1 (field 'add) (map (lambda (x y) ((field 'mul) x y)) lst1 lst2))
)


(define (to-maple-list ls)
  (define x "[")
  (for-each (lambda (y) (set! x (string-append x (number->string y) ","))) ls)
  (if (equal? #\, (string-ref x (- (string-length x) 1)))
      (string-set! x (- (string-length x) 1) #\])
      (set! x (string-append x "]"))
  )
  x
)


(define (to-maple-set ls)
  (define x "{")
  (for-each (lambda (y) (set! x (string-append x (number->string y) ","))) ls)
  (if (equal? #\, (string-ref x (- (string-length x) 1)))
      (string-set! x (- (string-length x) 1) #\})
      (set! x (string-append x "}"))
  )
  x
)

(define (to-maple-matrix m)
  (define x "[")
  (for-each (lambda (y) (set! x (string-append x y ","))) 
                                   (map (lambda (x) (to-maple-list x)) m)
  )
  (if (equal? #\, (string-ref x (- (string-length x) 1)))
      (string-set! x (- (string-length x) 1) #\])
      (set! x (string-append x "]"))
  )
  x 
)


(define (matrix-mul field m1 m2)
  (map (lambda (row1)
    (map (lambda (row2)
      (inner-product field row1 row2)
    ) (transpose m2))    
  ) m1)
)


;; only works for square matrix
(define (matrix-get-diagonal matrix)
  (map (lambda (row i)
    (list-ref row i)
  ) matrix (idx-interval matrix))
)


;; only works for square matrix
(define (matrix-set-diagonal! matrix lst)
  (for-each (lambda (row j x)
    (list-set! row j x)
  ) matrix (idx-interval lst) lst)
  matrix
)


(define (matrix-apply func matrix)
  (map (lambda (row)
    (map (lambda (x) (func x)) row)
  ) matrix)
)


(define (matrix-apply! func matrix)
  (for-each (lambda (row)
    (for-each (lambda (x j)
      (list-set! row j (func x))      
    ) row (idx-interval row) )
  ) matrix)
  matrix
)


(define (matrix-zero-out-lower! field matrix)
  (for-each (lambda (row i)
    (for-each (lambda (x j)
      (if (> i j) (list-set! row j (field 'zero)))      
    ) row (idx-interval row) )
  ) matrix (idx-interval matrix) )
  matrix
)


(define (matrix-zero-out-upper! field matrix)
  (for-each (lambda (row i)
    (for-each (lambda (x j)
      (if (> j i) (list-set! row j (field 'zero)))
    ) row (idx-interval row) )
  ) matrix (idx-interval matrix) )
  matrix
)


(define (matrix-swap-elements! matrix i j k l)
  (define ij (matrix-ref matrix i j))
  (define kl (matrix-ref matrix k l))
  (matrix-set! matrix i j kl)
  (matrix-set! matrix k l ij)
  matrix
)


(define (matrix-permutate! matrix plist) 
  (define m (list-copy matrix))
  (for-each (lambda (i j)
    (list-set! matrix i (list-ref m j))  
  ) (idx-interval plist) plist)
  matrix
)


(define (matrix-permutate matrix plist) 
  (map (lambda (i)
    (list-ref matrix i)
  ) plist)
)


(define (matrix-cat m1 m2)
  (map (lambda (r1 r2)
    (append r1 r2)
  ) m1 m2)
)


(define (row-reduce field matrix)
  (define pivot-row -1) 
  (define last-pivot-row -1) 
  (define pivot-columns '()) 
  (define tmp -1)
  (for-loop 0 (- (length (car matrix)) 1) (lambda (j)
    (set! pivot-row -1)
    (call/ec (lambda (return)
      (for-loop (+ last-pivot-row 1) (- (length matrix) 1) (lambda (i)
        (if (not ((field 'zero?) (matrix-ref matrix i j))) (begin (set! pivot-row i) (return 0)) )
      ))
    ))
    (if (> pivot-row -1) (begin
      (set! pivot-columns (cons j pivot-columns))
      (set! tmp ((field 'mul-inv) (matrix-ref matrix pivot-row j)))
      (matrix-row-op! matrix pivot-row (lambda (x) ((field 'mul) x tmp)) )
      (set! last-pivot-row (+ 1 last-pivot-row))
      (swap! pivot-row last-pivot-row matrix)
      (for-loop (+ last-pivot-row 1) (- (length matrix) 1) (lambda (i)
        (set! tmp (matrix-ref matrix i j))
        (if (not ((field 'zero?) tmp)) (begin
            (set! tmp ((field 'negate) ((field 'mul-inv) tmp)))
            (matrix-row-op! matrix i (lambda (x) ((field 'mul) x tmp)) )
            (list-set! matrix i 
              (map (lambda (x y) ((field 'add) x y)) 
                 (list-ref matrix i) (list-ref matrix last-pivot-row)
              )
            )
        ))
      ))
    ))
  ))
  (list last-pivot-row pivot-columns)
)


(define (rref field matrix)
  (let ((pivots (row-reduce field matrix))
        (tmp 0) 
       )
       (for-each (lambda (i j)
         (set! tmp ((field 'mul-inv) (matrix-ref matrix i j)))
         (matrix-row-op! matrix i (lambda (x) ((field 'mul) x tmp)))
         (if (> i 0) 
             (for-each (lambda (k)
               (set! tmp (matrix-ref matrix k j))
               (if (not ((field 'zero?) tmp)) (begin
                 (set! tmp ((field 'negate) ((field 'mul-inv) tmp)))
                 (matrix-row-op! matrix k (lambda (x) ((field 'mul) x tmp)) )
                 (list-set! matrix k
                   (map (lambda (x y) ((field 'add) x y))
                     (list-ref matrix k) (list-ref matrix i)
                   )
                 )
               ))
             ) (iota i (- i 1) -1) )
         )
       ) (iota (+ 1 (car pivots)) (car pivots) -1) (cadr pivots) )
       pivots
  )
)
(define (rank field matrix) (if (> (length matrix) 0) (+ 1 (car (rref field matrix))) 0))


(define (inverse field mat)
  (define matrix (list-copy mat))
  ;(define m (matrix-cat matrix (identity-matrix (length matrix))))
  (if (not (= (rank field matrix) (length matrix))) 'singular
      (let ((m (matrix-cat mat (identity-matrix (length mat))))
           )
           (rref field m)
           (map (lambda (row)
             (take-right (length mat) row)  
           ) m)
      )
  )
)

;; The number of nonzero rows in the reduced row echelon form of a matrix A 
;;     is called the rank
;; Theorem. Consider the m*n linear system Ax=b where M=<A|b> is the augmented matrix.
;; 1. If rank(A)=rank(M)=n then unique solution
;; 2. If rank(A)=rank(M)<n then infinite number of solution
;; 3. If rank(A)<rank(M)   then no solution
;;
;; determinant of triangular matrix is the product of the diagonal
;;
;; the inverse of a permutation matrix is simply its transpose
;;
;; use rref to find inverse (I A^-1)=rref(AI)
;;
;; If the determinant of a matrix is 0, the matrix is said to be singular
;;
;; not invertible <=> not full rank <=> singular <=> determinant=0
;;
;; http://www.math.uu.nl/people/bisselin/PSC/psc2_1.pdf
;;
;; all index starting with 1
;; lu-decomposition 
;; input A
;; output (determinant p L U)
;; pA=LU
;; this function can fail, check if p is null first 
;; only magnitude of determinant is correct (that is sign is not necessarily correct)
;; Note: LU decomposition is not unique: if A = LU, then A = L DD^(-1) U = (LD) (D^(-1)U) 
;; Note: full-rank seems to give the same LU as Maple does
;;
;; full-rank <=> determinant is not 0
;; det(AB)=det(A)det(B)
;; det(rA)=r^n det(A)
;; det(A)^-1 = det(A^-1) iff A is full-rank
;; 
(define (ludecomp field matrix)
  (call/ec (lambda (return)
    (let* ((n (length matrix))       ;; assume square matrix
           (p (interval 0 (- n 1)))  ;; permutation vector
           (r -1)                    ;; pivot row index 
           (tmp -1)
          )                   
          (for-loop 0 (- n 1) (lambda (k)
            (if ((field 'zero?) (matrix-ref matrix k k)) (begin ;; only swap if necessary
                (set! r (car (filter (lambda (x) (>= x k)) 
                               (map cdr (map-sort (field 'abs) (field 'greater-than) 
                                          (matrix-get-column matrix k)))
                             )))
                (swap! k r p)       ;; record rows swap
                (swap! k r matrix)  ;; rows swap
            ))
            (if ((field 'zero?) (matrix-ref matrix k k)) (return (list 0 '() '() '()))) ;; abort
            (set! tmp ((field 'mul-inv) (matrix-ref matrix k k)))
            (for-loop (+ k 1) (- n 1) (lambda (i)
              (matrix-set! matrix i k ((field 'mul) (matrix-ref matrix i k) tmp))
            ))
            (for-loop (+ k 1) (- n 1) (lambda (i)
            (for-loop (+ k 1) (- n 1) (lambda (j)
              (matrix-set! matrix i j 
                ((field 'add) (matrix-ref matrix i j)
                  ((field 'negate) ((field 'mul) (matrix-ref matrix i k) 
                                                 (matrix-ref matrix k j)))
                )
              )
            ))))
          ))
          ;; return value (det p L U)
          (let ((L (list-copy matrix))
                (U (list-copy matrix))
                (det (foldr1 (field 'mul) (matrix-get-diagonal matrix)))
               )
               (matrix-zero-out-upper! field L)
               (matrix-set-diagonal! L (map (lambda (i) (field 'one)) (interval 1 n)))
               (matrix-zero-out-lower! field U)
               (list det p L U)
          )
    ) 
  ))
)


(define (inc x) (if (list? x) (map (lambda (x) (+ x 1)) x) (+ x 1)))
(define (dec x) (if (list? x) (map (lambda (x) (- x 1)) x) (- x 1)))


;; x^d mod n
(define (powermod x d n)
  (cond ((= d 0) 1)
        ((even? d) (mod (square (powermod x (/ d 2) n)) n))
        (else (mod (* x (powermod x (- d 1) n)) n))
  ) 
)
(define (powermod-table n)
  (map (lambda (x) (map (lambda (y) (powermod x y n)) (interval 1 (- n 1)))
       ) (interval 1 (- n 1))))  


;; (n>=1) n=x^y for some y>=2
(define (perfect-power-of? n x) ;; n=x^y for some y>=2
   (if (< n 1) (error "n must >=1"))
   (if (= n 1) #t
      (if (integer? (/ n x)) (perfect-power-of? (/ n x) x)
          #f
      )
   )
)


(define (log2 n)
  (define r 0)
  (let iter ()
       (if (>= n 2) 
           (begin (set! n (/ n 2)) 
                  (set! r (+ r 1))
                  (iter)
           )
       )
  )
  r
)


(define (factor n)
  (if (= n 0) (list 0)
      (if (< n 0) (cons -1 (factor (abs n)))
          (if (= n 1) (list 1) 
              (call/ec (lambda (return)
                (for-loop 2 n (lambda (x)
                   (if (= 0 (remainder n x)) (return (cons x (factor (quotient n x))))) 
                ))
              ))
          )
      )
  )
)


;;;;;;;;;;;;;;;;;;;;; 
;; polynomial library 
;;
;; http://www.geocities.com/CapeCanaveral/Hall/3131/
;;

(define (mdeg monomial)
  (if (null? monomial) (error "null monomial") (foldr1 + monomial))
)


(define (lex l1 l2)
  (if (or (null? l1) (null? l2)) (error "lex: empty term")
    (call/ec (lambda (return)
      (let ((r (map - l1 l2)))
        (for-each (lambda (x) 
          (if (not (zero? x)) 
              (if (positive? x) (return #t) (return #f))
          )
        ) r)
      )
    #f))
  )
)


(define (rlex l1 l2)
  (if (or (null? l1) (null? l2)) (error "rlex: empty term")
    (call/ec (lambda (return)
      (let ((r (map - l1 l2)))
        (for-each (lambda (x) 
          (if (not (zero? x)) 
              (if (negative? x) (return #t) (return #f))
          )
        ) (reverse r) )
      )
    #f))
  )
)


(define (grlex a b)
  (if (or (null? a) (null? b)) (error "grlex: empty term")
    (call/ec (lambda (return)
      (if (> (mdeg a) (mdeg b)) (return #t)) 
      (if (and (= (mdeg a) (mdeg b)) (lex a b)) (return #t))
    #f))
  )
)


(define (grevlex a b)
  (if (or (null? a) (null? b)) (error "grevlex: empty term")
    (call/ec (lambda (return)
      (if (> (mdeg a) (mdeg b)) (return #t))
      (if (and (= (mdeg a) (mdeg b)) (rlex a b)) (return #t))
    #f))
  )
)


(define (kth-elim kth a b)
  (define sa (mdeg (take kth a)))
  (define sb (mdeg (take kth b)))
  (if (or (null? a) (null? b)) (error "kth-elim-grevlex: empty term")
    (if (> sa sb) #t
        (if (= sa sb) (grevlex a b) #f)
    )
  )
)


(define kth1 (lambda (x y) (kth-elim 1 x y)) )
(define kth2 (lambda (x y) (kth-elim 2 x y)) )
(define kth3 (lambda (x y) (kth-elim 3 x y)) )




(define (m-posdiv? monomial1 monomial2)
  (= (length (filter negative? (map (lambda (x y) (- x y)) monomial1 monomial2))) 0)
)
(define (m-alldiv? monomial1 monomial2) #t)



;;
;; polynomial property examples
;;
(define (lex-rational term-map)
  (lambda (prop)
    (cond ((eq? prop 'morder)                lex )  ;; monomial ordering function
          ((eq? prop 'field)      rational-field )  ;; coefficient field
          ((eq? prop 'mdiv?)           m-posdiv? )  ;; monomial divide?
          ((eq? prop 'tmap)             term-map )  ;; term strings map
          (else       (error "no such property") )
    )
  )
)
(define (grevlex-rational term-map)
  (lambda (prop)
    (cond ((eq? prop 'morder)            grevlex )  ;; monomial ordering function
          ((eq? prop 'field)      rational-field )  ;; coefficient field
          ((eq? prop 'mdiv?)           m-posdiv? )  ;; monomial divide?
          ((eq? prop 'tmap)             term-map )  ;; term strings map
          (else       (error "no such property") )
    )
  )
)
(define (lex-Zp term-map p)
  (lambda (prop)
    (cond ((eq? prop 'morder)                lex )  ;; monomial ordering function
          ((eq? prop 'field)        (Zp-field p) )  ;; coefficient field
          ((eq? prop 'mdiv?)           m-posdiv? )  ;; monomial divide?
          ((eq? prop 'tmap)             term-map )  ;; term strings map
          (else       (error "no such property") )
    )
  )
)
(define lex-rational-xyz (lex-rational (list "x" "y" "z")))
(define grevlex-rational-xyz (grevlex-rational (list "x" "y" "z")))



(define (last-char string) (string-ref string (- (string-length string) 1)))


(define (chop string n) (substring string 0 (- (string-length string) n)))


;;
;; polynomial data structure
;; (list (list coeff deg1 deg2 deg3 ...) ...)
;;
(define (term-string pp term)
  (define ret (string-append "(" (((pp 'field) 'str) (car term)) "*" ))
  (define term-map- (map (lambda (i) (list-ref (pp 'tmap) (- i 1))) 
                    (interval 1 (length (cdr term)))))
  (if (and (eq? ((pp 'field) 'one) (car term)) 
           (not (all-zero? ((pp 'field) 'zero?) (cdr term)))
      ) 
      (set! ret "(")
  )
  (for-each (lambda (sym deg) 
    (if (not (zero? deg))
        (if (= deg 1)
            (set! ret (string-append ret sym "*" ))
            (set! ret (string-append ret sym "^" (number->string deg) "*" ))
        )
    )
  ) term-map- (cdr term) ) 
  (if (eq? #\* (last-char ret))
      (string-set! ret (- (string-length ret) 1) #\) )
      (set! ret (string-append ret ")"))
  )
  ret
)


(define (poly-string pp polynomial)
  (define ret "")
  (if (null? polynomial) (((pp 'field) 'str) ((pp 'field) 'zero)) (begin
    (for-each (lambda (term) 
      (set! ret (string-append ret (term-string pp term) "+"))
    ) polynomial)
    (if (eq? #\+ (last-char ret)) (chop ret 1) ret )
  ))
)


(define (polys-string pp polys)
  (define ret "[")
  (for-each (lambda (poly) 
    (set! ret (string-append ret (poly-string pp poly) ","))
  ) polys)
  (if (eq? #\, (last-char ret)) 
      (string-set! ret (- (string-length ret) 1) #\])
      (set! ret (string-append ret "]"))
  )
  ret
)


(define (term-list-string pp term)
  (define ret "(")
  (set! ret (string-append ret (((pp 'field) 'str) (car term)) " "))
  (for-each (lambda (exponent)
    (set! ret (string-append ret (number->string exponent) " "))
  ) (cdr term))  
  (string-set! ret (- (string-length ret) 1) #\))
  ret
)


(define (poly-list-string pp poly)
  (define ret "(")
  (for-each (lambda (term)
    (set! ret (string-append ret (term-list-string pp term)))
  ) poly)
  (string-append ret ")")
)


(define (polys-list-string pp polys)
  (append 
    (cons "# "
      (map (lambda (poly) 
        (string-append "# " (poly-list-string pp poly) " ")
      ) polys )
    ) 
    (list "# ")
  )
)


(define (poly-sort! pp poly)
  (define p (insertion-sort (lambda (x y) ((pp 'morder) (cdr x) (cdr y))) poly))
  (for-each (lambda (term i)
    (list-set! poly i (list-ref p i))
  ) poly (idx-interval poly))
  p
)


(define (LT pp poly)
  (car (poly-sort! pp poly))
)
(define (LC pp poly)
  (car (car (poly-sort! pp poly)))
)


;; term1 add "onto" term2
(define (TaddT field term1 term2)
  (if (equal? (cdr term1) (cdr term2))
      (let ((coeffsum ((field 'add) (car term1) (car term2)))) 
           (if ((field 'zero?) coeffsum) 0 (cons coeffsum (cdr term2))) ;; 0 denote 0 term
      )
      term2
  )
)


(define (TmulT field term1 term2)
  (if (or ((field 'zero?) (car term1)) ((field 'zero?) (car term2))) 0
    (cons ((field 'mul) (car term1) (car term2)) 
          (map (lambda (x y) (+ x y)) (cdr term1) (cdr term2))
    )
  )
)


(define (TmulP field term poly)
  (delete eq? 0 (map (lambda (x) (TmulT field x term)) poly))
)


;; term add "onto" poly
(define (TaddP field term poly)
  (if ((field 'zero?) (car term)) poly 
    (let ((ret (delete eq? 0 (map (lambda (x) (TaddT field term x)) poly))))
         (if (equal? ret poly) (cons term poly) ret)
    )
  )
)


(define (scan-out-0-terms field poly)
  (delete eq? 0 (map (lambda (term) (if ((field 'zero?) (car term)) 0 term)) 
                  (delete eq? 0 poly))
  )
)


;; three levels of 0: 
;; 1. 0-polynomial '()
;; 2. 0-term        0
;; 3. 0-coeff      (0 ...)
(define (poly-add field poly1 poly2)
  (let ((ret (scan-out-0-terms field poly2)))
       (for-each (lambda (term) (set! ret (TaddP field term ret))) 
                   (scan-out-0-terms field poly1))
       ret
  )
)


(define (poly-mul field poly1 poly2)
  (if (or (null? poly1) (null? poly2)) '()
      (foldr1 (lambda (x y) (poly-add field x y))
        (map (lambda (x) (TmulP field x poly2)) poly1)
      )
  )
)


(define (poly-negate field poly)
  (map (lambda (term) (cons ((field 'negate) (car term)) (cdr term))) poly)
)


(define (poly-sub field poly1 poly2)
  (poly-add field poly1 (poly-negate field poly2))
)


(define (TdivT? pp term1 term2)
  ((pp 'mdiv?) (cdr term1) (cdr term2))
)


(define (TdivT field term1 term2)
  (cons ((field 'mul) (car term1) ((field 'mul-inv)(car term2)) )  
        (map (lambda (x y) (- x y)) (cdr term1) (cdr term2)) 
  )
)


;;
;; f = q g + r
;; http://www.geocities.com/CapeCanaveral/Hall/3131/
;; Input: f, g
;; Output: q, r
;; q := 0; r := f
;; WHILE (r<>0 AND LT(g) divides LT(r)) DO
;;    q := q + LT(r)/LT(g)
;;    r := r - (LT(r)/LT(g))g
;; note: at every step f = q g + r holds
;;
(define (poly-div pp f g)
  (define q '())
  (define r f)
  (define LTg (LT pp g))
  (define LTr (list 0))     ;; use assign and car to set and get
  (define LTr/LTg (list 0))
  ;(print (string-append "# poly-div f:" (poly-list-string pp f) )) 
  ;(print (string-append "# poly-div g:" (poly-list-string pp g) )) 
  (while (and (not (null? r)) (TdivT? pp (assign LTr (LT pp r)) LTg))
    (set! LTr/LTg (TdivT (pp 'field) (car LTr) LTg))
    ;; because LTg is fixed, and LTr is strictly decreasing, so you can use cons
    (set! q (cons (assign LTr/LTg (TdivT (pp 'field) (car LTr) LTg)) q)) 
    (set! r (poly-sub (pp 'field) r (TmulP (pp 'field) (car LTr/LTg) g)))
  )
  (list r q)
)

;; convert a number to an univariate polynomial
(define (num-to-poly n b)
  (define l (convert-to-base n b))
  (if (null? l) l
    (scan-out-0-terms (lambda (x) zero?)
      (map (lambda (c d) (list c d))
           l (interval 0 (- (length l) 1))
      )         
    )
  )           
)  


(define (monic? pp poly) (((pp 'field) 'one?) (LC pp poly)))


;; prepare polynomial (merge terms)
(define (poly-prep field poly) (poly-add field poly '()))


(define (with-duplicate-monomials? poly)
  (let* ((monos (map cdr poly))
         (len1 (length monos))
         (tmp (list->set equal? monos))
         (len2 (length tmp)) 
        )
        (if (not (= len1 len2)) #t #f)
  )
)


;; general division algorithm
;; http://www.geocities.com/CapeCanaveral/Hall/3131/
;; input: (f g1 g2 ...)
;;   f=q1g1+q2g2+...+r
;; output: (r q1 q2)
(define (poly-gdiv pp fG)
  (let* ((f (car fG))
         (G (cdr fG))
         (LTG (map (lambda (p) (LT pp p)) G))
         (LTf 0)
         (size-of-G (length G))
         (Q (list->vector (map (lambda (i) '()) (interval 1 size-of-G))))
         (r '()) (i 1) (dividing #t) (tmp 0)
        ) 
        (while (not (null? f))
          (set! i 0)
          (set! dividing #t)
          (set! LTf (LT pp f))
          (while (and (< i size-of-G) dividing)
            (if (TdivT? pp LTf (list-ref LTG i)) (begin
                (set! tmp (TdivT (pp 'field) LTf (list-ref LTG i))) 
                (vector-set! Q i (poly-add (pp 'field) (vector-ref Q i) (list tmp) ) )
                (set! f (poly-sub (pp 'field) f (TmulP (pp 'field) tmp (list-ref G i))))
                (set! dividing #f)
            ) (set! i (+ i 1)) )
          )
          (if dividing (begin  ;; nothing was done the previous while loop
              (set! r (append r (list LTf))) 
              (set! f (delete equal? LTf f))
              ;(if (with-duplicate-monomials? r) (error "bug?"))
          ))
        )
        (cons r (vector->list Q))
  )
)


(define (normalf pp fG)
  (let* ((f (car fG))
         (G (cdr fG))
         (LTG (map (lambda (p) (LT pp p)) G))
         (LTf 0)
         (size-of-G (length G))
         ;(Q (list->vector (map (lambda (i) '()) (interval 1 size-of-G))))
         (r '()) (i 1) (dividing #t) (tmp 0)
        ) 
        (while (not (null? f))
          (set! i 0)
          (set! dividing #t)
          (set! LTf (LT pp f))
          (while (and (< i size-of-G) dividing)
            (if (TdivT? pp LTf (list-ref LTG i)) (begin
                (set! tmp (TdivT (pp 'field) LTf (list-ref LTG i))) 
                ;(vector-set! Q i (poly-add (pp 'field) (vector-ref Q i) (list tmp) ) )
                (set! f (poly-sub (pp 'field) f (TmulP (pp 'field) tmp (list-ref G i))))
                (set! dividing #f)
            ) (set! i (+ i 1)) )
          )
          (if dividing (begin  ;; nothing was done the previous while loop
              (set! r (append r (list LTf))) 
              (set! f (delete equal? LTf f))
              ;(if (with-duplicate-monomials? r) (error "bug?"))
          ))
        )
        r ;(cons r (vector->list Q))
  )
)


(define (poly-gdiv-verify pp fG)
  (define rQ (poly-gdiv pp fG))
  (let* ((f (car fG))
         (G (cdr fG))
         (r (car rQ))
         (Q (cdr rQ))
        )
        (poly-sub (pp 'field) f
          (foldr1 (lambda (x y) (poly-add (pp 'field) x y))
            (cons r (map (lambda (x y) (poly-mul (pp 'field) x y)) Q G))
          )
        )
  ) 
)


;; polynomial set monomial exponent gcd
(define (polys-mgcd F)
  (map (lambda (lst) (apply min lst)) (transpose (map cdr (flatten-1 F))))
)


(define (take-out-mgcd mgcd) 
  (lambda (poly)
    (map (lambda (term) (cons (car term) (map - (cdr term) mgcd))) poly) 
  )
)


(define (put-back-mgcd mgcd) 
  (lambda (poly)
    (map (lambda (term) (cons (car term) (map + (cdr term) mgcd))) poly) 
  )
)


(define (coeffs poly) (map (lambda (term) (car term)) poly))
(define (LCM monomial1 monomial2) (map (lambda (x y) (max x y)) monomial1 monomial2))
(define (zip func lst1 lst2) (map (lambda (x y) (func x y)) lst1 lst2))


(define (poly-scale field scalar poly)
  (map (lambda (term) (cons ((field 'mul) scalar (car term)) (cdr term))) poly)
)


;; S-polynomial
;; J = LCM(LM(f),LM(g))
;; S(f,g) = f LC(g)LC(f) J/LT(f) - g LC(g)LC(f) J/LT(g)
(define (s-poly pp f g)
  (let* ((LTf (LT pp f))
         (LTg (LT pp g))
         (J (LCM (cdr LTf) (cdr LTg)) )
         (f-term (cons (car LTg) (zip - J (cdr LTf))))
         (g-term (cons (car LTf) (zip - J (cdr LTg))))
        )
        (poly-sub (pp 'field) (TmulP (pp 'field) f-term f)    
                              (TmulP (pp 'field) g-term g))
  )
)


(define (polys->rref pp S)
 (if (null? (car S)) '(())
  (let* ((sorted-monos (insertion-sort (lambda (x y) ((pp 'morder) x y))
           (list->set equal? (map cdr (flatten-1 S)))
         ))
         (mat (map (lambda (p)  
                (map (lambda (m)
                  (call/ec (lambda (return)
                    (for-each (lambda (t)
                      (if (equal? m (cdr t)) (return (car t)))
                    ) p)
                  ((pp 'field) 'zero)))
                ) sorted-monos)
              ) S) 
         )
         (res (rref (pp 'field) mat))
         (rrefpolys (map (lambda (i)
                      (scan-out-0-terms (pp 'field)
                        (map (lambda (c j)
                          (cons c (list-ref sorted-monos j))
                        ) (list-ref mat i) (interval 0 (- (length (car mat)) 1)))
                      )
                    ;) (interval 0 (car res)))
                    ) (iota (+ 1 (car res)) (car res) -1))
         )
        )
        rrefpolys
  )
 )
)


(define (grobner? pp S)
  (define len (length S))
  (call/ec (lambda (return)
    (for-loop 0 (- len 1) (lambda (i)
    (for-loop 0 (- len 1) (lambda (j)
      (if (not (= i j))
          (if (not (null?
                (normalf pp (cons (s-poly pp (list-ref S i) (list-ref S j)) S) )
              )) (return #f)
          )
      )
    ))))
  #t))
)

;; generate indices set for n elements
(define (gen-indices-set n)
  (define ret '())
  (for-loop 0 (- n 1)       (lambda (i)
  (for-loop (+ i 1) (- n 1) (lambda (j)
    (set! ret (append ret (list (cons i j)) ))
  ))))
  ret
)


;; has the side-effect of sorting the polynomial (because of LC)
(define (poly-make-monic pp poly)
  (if (null? poly) poly 
    (poly-scale (pp 'field) (((pp 'field) 'mul-inv) (LC pp poly)) poly)
  )
)


;; NOTE: do poly-sort! on each poly before calling. 
;; "smaller" poly first
(define (polys-sort pp polys)
  (insertion-sort 
    (lambda (p1 p2) 
      (if (null? p1) #t
        (if (null? p2) #f
          ;(not ((pp 'morder) (cdr (car p1)) (cdr (car p2))))
          (if ((pp 'morder) (cdr (car p1)) (cdr (car p2))) #f
              (if (equal? (cdr (car p1)) (cdr (car p2))) 
                  (polys-sort pp (list (cdr p1) (cdr p2)))
                  #t
              )
          ) 
        )
      )
    ) 
  polys)
)


(define (poly-equal? field p q)  (null? (poly-sub field p q))  )
(define (poly-square field p) (poly-mul field p p) )
(define (poly-cube field p) (poly-mul field p (poly-square field p)) )



;;
;; Input: A polynomial set F = {f1,f2,...,fs} that generates an ideal I
;; Output: A Grobner basis G = {g1,g2,...,gt} that generates the same I
;;
(define (naive-buchberger pp F)
  (let* ((G (polys-sort pp (map (lambda (p) (poly-make-monic pp p)) F)  )) ;; helps a lot 
         (lenG (length G))
         (M (gen-indices-set lenG))
         (ij 0) (s 0) (r 0) 
        )
        (while (not (null? M))
          (set! ij (car M)) (set! M (cdr M)) ;; dequeue
          (let* ((LMp (cdr (car (list-ref G (car ij))))) 
                 (LMq (cdr (car (list-ref G (cdr ij)))))
                 (LCMpq (LCM LMp LMq))
                 (tmp (interval 0 (- (min (car ij) (cdr ij)) 1) ))
                 (cr2 #f)
                 (tmp2 (for-each (lambda (i) 
                         (if ((pp 'mdiv?) LCMpq (cdr (car (list-ref G i))))
                             (set! cr2 #t)
                         ) 
                       ) tmp))   ;; Criterion 2, helps with inconsistent systems 
                )
                (if (not (equal? (zip + LMp LMq) LCMpq)) (begin ;; Criterion 1
                  (if (not cr2) (begin                          ;; Criterion 2
                    (set! s (s-poly pp (list-ref G (car ij)) (list-ref G (cdr ij))))
                    (set! r (normalf pp (cons s G)))
                    (if (not (null? r)) (begin
                        (set! r (poly-make-monic pp r))  
                        (set! G (append G (list r)))
                        (for-loop 0 (- lenG 1) (lambda (i)
                          (set! M (append M (list (cons i lenG))))
                        ))
                        (set! lenG (+ 1 lenG))
                    ))
                ))))
          )
        )
        (list->set equal? G) ;; return value, all polynomials monic and sorted
  )
)


;; http://www.lifl.fr/evenements/publications/2001-07.ps
(define (buchberger pp L)
  (if (null? L) (error "for pete sake, need at least one polynomial")
    (let* ((L (map (lambda (p) (list (car p) p)) 
                ;(polys-sort pp (map (lambda (p) (poly-make-monic pp p)) L))
                (polys->rref pp L)
              ))                             ;; (list (list LTp p) ...)
           (A (list 0))
           (D '())                           ;; (list (list i j LTs s sugs) ...)
           (P (interval 1 (- (length L) 1)))
           (getA (lambda () (map (lambda (i) (second (list-ref L i))) A)))
           (scmp? (lambda (s1 s2) (if (or (and (= (car s1) (car s2)) (= (cadr s1) (cadr s2)) )
                                          (and (= (car s1) (cadr s2)) (= (cadr s1) (car s2)) )
                                      ) #t #f)))
           (p 0) (usedD '()) 

           (s-poly (lambda (LTf LTg f g)
             (let* ((J (LCM (cdr LTf) (cdr LTg)) )
                    (f-term (cons (car LTg) (zip - J (cdr LTf))))
                    (g-term (cons (car LTf) (zip - J (cdr LTg))))
                   )
                   (poly-sub (pp 'field) (TmulP (pp 'field) f-term f)    
                                         (TmulP (pp 'field) g-term g))
             )))
           (normalf (lambda (f) ;; note: A and L use global
             (let* ((LTf 0)
                    (LTG (map (lambda (i) (first  (list-ref L i))) A))
                    (G   (map (lambda (i) (second (list-ref L i))) A))
                    (size-of-G (length G)) 
                    (r '()) (i 1) (dividing #t) (tmp 0) 
                   ) 
                   (while (not (null? f)) 
                     (set! i 0)                         
                     (set! dividing #t)
                     (set! LTf (LT pp f))
                     (while (and (< i size-of-G) dividing)
                       (if (TdivT? pp LTf (list-ref LTG i)) (begin
                           (set! tmp (TdivT (pp 'field) LTf (list-ref LTG i))) 
                           (set! f (poly-sub (pp 'field) f 
                                     (TmulP (pp 'field) tmp (list-ref G i))))
                           (set! dividing #f)           
                       ) (set! i (+ i 1)) )
                     )
                     (if dividing (begin  ;; nothing was done the previous while loop
                         (set! r (append r (list LTf))) 
                         (set! f (delete equal? LTf f))
                     ))
                   )
                   r
             )))
          )

          (while (or (not (null? D)) (not (null? P)) )
            (if (null? D) 
                (begin
                  (set! p (car P))
                  (set! P (cdr P))
                  (set! p (normalf (second (list-ref L p)) )) ;; index -> poly
                  (set! p (poly-make-monic pp p))
                ) 
                (begin 
                  (set! p (car D))
                  (set! D (cdr D))
                  ;; criterion 2 ;; don't apply if use rref
                  (set! usedD (cons p usedD))
                  (let* ((cr2 #f)
                         (i (first p))
                         (j (second p))
                        )
                        (call/ec (lambda (return)
                          (for-each (lambda (k)
                            (if (or (= k i) (= k j)) '() ;; do nothing
                                (if (and (contains? scmp? (list i k) usedD)
                                         (contains? scmp? (list j k) usedD)
                                         ((pp 'mdiv?) (LCM (cdr (first (list-ref L i))) 
                                                           (cdr (first (list-ref L j))))
                                                      (cdr (first (list-ref L k))))
                                    )
                                    (begin (set! cr2 #t) (return 0))
                                )
                            )
                          ) A)
                        ))
                        (if (not cr2) (begin
                          (set! p (normalf (fourth p) )) 
                          ;(if (null? p) (print "# s-poly reduced zero"))
                          (set! p (poly-make-monic pp p))
                        ) (set! p '()))
                  )     
                )
            )
            (let* ( (A- '()) (D- '()) (ok 0) )
                  (if (not (null? p)) (begin
                      (for-each (lambda (q)
                        (if (not (TdivT? pp (first (list-ref L q)) (car p)))
                            (set! A- (cons q A-))
                        )
                      ) A)
                      (set! L (append L (list (list (car p) p))))
                      (set! A- (cons (- (length L) 1) A-)) ;; new A all done
                      (set! p (- (length L) 1))
                      (set! ok #f)
                      (for-each (lambda (q)
                        (if (or (TdivT? pp (first (list-ref L p)) (first (list-ref L q)))
                                (TdivT? pp (first (list-ref L q)) (first (list-ref L p)))
                            ) ;; same as criterion 1 
                            (set! ok #t)
                            (if (and (contains? = p A-) (contains? = q A-)) ;; criterion 3
                                (set! ok #t)
                            )
                        )
                        (if ok 
                            (let* ((s (s-poly (first (list-ref L p)) (first (list-ref L q))
                                        (second (list-ref L p)) (second (list-ref L q)) ))
                                   (s (poly-make-monic pp s))
                                  )
                                  (if (not (null? s)) (set! D- (cons (list p q (car s) s) D-)))
                            )
                        )
                      ) A)
                      (set! A (insertion-sort (lambda (x y) 
                                    (not ((pp 'morder) (cdr (first (list-ref L x))) 
                                                       (cdr (first (list-ref L y))) ))
                              ) A-))
                      (set! D (insertion-sort (lambda (x y) 
                                    (not ((pp 'morder) (cdr (third x)) (cdr (third y)) ))
                              ) (append D D-)))
                  ))
            )
          )
          ;; return value
          (getA)
    )
  )
)


;; this should only be applied on grobner, ie all s-polys -> 0 (mod G)
(define (g-reduce pp G)
  (if (< (length G) 2) G
      (let* (
             (G-  (map (lambda (f) (normalf pp (cons f (delete equal? f G)))) G))
             (G-- (delete eq? '() G-))
             (tmp '())
            )
            (for-each (lambda (x y)
              (if (null? y) (set! tmp (cons (normalf pp (cons x G--)) tmp)) )
            ) G G-)
            (set! tmp (delete eq? '() tmp))
            (if (not (null? tmp)) (begin
                ;(print "# it happens")
                (if (equal? (reverse tmp) G) G (begin
                    ;(print "# it happenss") 
                    (g-reduce pp (append tmp G--))
                ))
            ) G--)
      )
  )
)



;(define (gbasis pp S) (g-reduce pp (naive-buchberger pp (polys->rref pp S))))
(define (gbasis pp S) (g-reduce pp (buchberger pp S)))

(define (permu lst) (bag-of (map (lambda (i) (number-between 0 i)) lst)))


(define (polys-big-small pp S)
  (let ((big (car (flatten-1 S)))
        (small (car (flatten-1 S)))
       )
       (for-each (lambda (t)
         (if ((pp 'morder) (cdr t) (cdr big))
             (set! big t) 
             (if ((pp 'morder) (cdr t) (cdr small)) '()
                 (set! small t)
             )
         )
       ) (flatten-1 S) )
       (list big small)
  )
)
